'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const deepEqual = require("fast-deep-equal");
const querystring = require("querystring");
const xhr = require("request");
const produceURL = 'https://www.varsnap.com/api/snap/produce/';
const consumeURL = 'https://www.varsnap.com/api/snap/consume/';
const produceTrialURL = 'https://www.varsnap.com/api/trial/produce/';
exports.version = 'v1.3.1';
exports.configKeys = {
    varsnap: 'varsnap',
    env: 'env',
    producerToken: 'producerToken',
    consumerToken: 'consumerToken',
    logger: undefined,
};
let Producers = [];
let Consumers = [];
class Util {
    static getConfig(config, configKey) {
        if (!config) {
            return '';
        }
        let val = config[configKey];
        if (!val) {
            return '';
        }
        val = val.toString().toLowerCase();
        return val;
    }
    static getSignature(targetFunc) {
        let signature = targetFunc.toString().split('{')[0];
        signature = signature.replace(/^\s+|\s+$/g, '');
        return 'javascript.' + exports.version + '.' + signature;
    }
    static ajax(url, requestData) {
        const options = {
            method: 'POST',
            uri: url,
            body: querystring.stringify(requestData),
            headers: { 'Content-Type': 'application/x-www-form-urlencoded' },
        };
        const ajaxPromise = new Promise((resolve, reject) => {
            xhr(options, (err, resp, body) => {
                if (err) {
                    return reject(err);
                }
                const responseCode = Math.round(resp.statusCode / 100);
                if (responseCode != 2) {
                    return reject(new Error('Status: ' + resp.statusCode));
                }
                body = JSON.parse(body);
                return resolve(body);
            });
        });
        return ajaxPromise;
    }
    static debug(data) {
        Util.logger.debug(data);
    }
    static log(data) {
        Util.logger.log(data);
    }
    static error(data) {
        Util.logger.error(data);
    }
    static resetProducers() {
        Producers = [];
    }
    static resetConsumers() {
        Consumers = [];
    }
    static getProducers() {
        return Producers;
    }
    static getConsumers() {
        return Consumers;
    }
    static limitString(x) {
        x = String(x);
        const limit = 30;
        const ellipsis = '...';
        if (x.length <= limit) {
            return x;
        }
        return x.substring(0, 30 - ellipsis.length) + ellipsis;
    }
}
exports.Util = Util;
Util.logger = console; // eslint-disable-line no-console
class Producer {
    constructor(targetFunc, config) {
        this.targetFunc = targetFunc;
        this.config = config;
    }
    isEnabled() {
        if (Util.getConfig(this.config, exports.configKeys.varsnap) !== 'true') {
            return false;
        }
        if (Util.getConfig(this.config, exports.configKeys.env) !== 'production') {
            return false;
        }
        if (!Util.getConfig(this.config, exports.configKeys.producerToken)) {
            return false;
        }
        return true;
    }
    static serialize(data) {
        return JSON.stringify(data);
    }
    produce(args, output) {
        if (!this.isEnabled()) {
            return;
        }
        Util.debug('Sending call to Varsnap');
        const data = {
            producer_token: Util.getConfig(this.config, exports.configKeys.producerToken),
            signature: Util.getSignature(this.targetFunc),
            inputs: Producer.serialize(args),
            prod_outputs: Producer.serialize(output),
        };
        Util.ajax(produceURL, data).catch((err) => {
            if (err.statusCode === 404) {
                Util.error('Cannot produce to Varsnap; check your producer token');
            }
            else {
                Util.error(err.error);
            }
        });
    }
}
exports.Producer = Producer;
class Consumer {
    constructor(targetFunc, config) {
        this.targetFunc = targetFunc;
        this.config = config;
    }
    isEnabled() {
        if (Util.getConfig(this.config, exports.configKeys.varsnap) !== 'true') {
            return false;
        }
        if (Util.getConfig(this.config, exports.configKeys.env) !== 'development') {
            return false;
        }
        if (!Util.getConfig(this.config, exports.configKeys.consumerToken)) {
            return false;
        }
        return true;
    }
    static deserialize(data) {
        if (data === '') {
            return undefined;
        }
        return JSON.parse(data);
    }
    consumeOnce() {
        if (!this.isEnabled()) {
            return;
        }
        const data = {
            consumer_token: Util.getConfig(this.config, exports.configKeys.consumerToken),
            signature: Util.getSignature(this.targetFunc),
        };
        return Util.ajax(consumeURL, data).then((response) => {
            if (!response || response.status !== 'ok') {
                return false;
            }
            let allMatches = true;
            for (const snap of response.results) {
                const inputs = Consumer.deserialize(snap['inputs']);
                const prodOutputs = Consumer.deserialize(snap['prod_outputs']);
                let localOutputs = undefined;
                try {
                    localOutputs = this.targetFunc(...inputs);
                }
                catch (err) {
                    localOutputs = err;
                }
                const matches = deepEqual(prodOutputs, localOutputs);
                this.reportCentral(snap, localOutputs, matches).then((trialURL) => {
                    this.reportLog(snap['id'], inputs, prodOutputs, localOutputs, matches, trialURL);
                });
                allMatches = allMatches && matches;
            }
            return allMatches;
        }).catch((error) => {
            Util.error('Cannot contact Varsnap: ' + error);
            return false;
        });
    }
    reportCentral(snap, localOutputs, matches) {
        const data = {
            'consumer_token': Util.getConfig(this.config, exports.configKeys.consumerToken),
            'snap_id': snap.id,
            'test_outputs': Producer.serialize(localOutputs),
            'matches': matches,
        };
        return Util.ajax(produceTrialURL, data).then((response) => {
            if (!response || response.status !== 'ok') {
                return false;
            }
            const trialURL = response['trial_url'];
            return trialURL;
        }).catch((error) => {
            Util.error('Cannot upload test results to Varsnap: ' + error);
            return false;
        });
    }
    reportLog(snapID, inputs, prodOutputs, localOutputs, matches, trialURL) {
        const functionName = Util.getSignature(this.targetFunc);
        Util.log('Testing with Varsnap uuid:     ' + snapID);
        Util.log('Function:                      ' + functionName);
        Util.log('Function input args:           ' + Util.limitString(inputs));
        Util.log('Production function outputs:   ' + Util.limitString(prodOutputs));
        Util.log('Your function outputs:         ' + Util.limitString(localOutputs));
        Util.log('Matching outputs:              ' + matches);
        Util.log('Trial URL:                     ' + trialURL);
        Util.log('');
    }
}
exports.Consumer = Consumer;
function varsnap(func) {
    if (varsnap.config === undefined) {
        console.error('Varsnap has no config; disabling.');
        return func;
    }
    if (varsnap.config.logger) {
        Util.logger = varsnap.config.logger;
    }
    const producer = new Producer(func, varsnap.config);
    const consumer = new Consumer(func, varsnap.config);
    const wrapped = function (...args) {
        let result = '';
        try {
            result = func(...args);
        }
        catch (err) {
            producer.produce(args, err);
            throw (err);
        }
        producer.produce(args, result);
        return result;
    };
    wrapped.producer = producer;
    wrapped.consumer = consumer;
    Producers.push(producer);
    Consumers.push(consumer);
    return wrapped;
}
exports.varsnap = varsnap;
varsnap.config = exports.configKeys;
varsnap.version = exports.version;
function runTests() {
    return __awaiter(this, void 0, void 0, function* () {
        const consumers = Util.getConsumers();
        let status = true;
        for (const consumer of consumers) {
            const testStatus = yield consumer.consumeOnce();
            status = status && testStatus;
        }
        return status;
    });
}
exports.runTests = runTests;
varsnap.runTests = runTests;
